@@ -106,16 +106,13 @@ func parseCaddyfileURI(h httpcaddyfile.Helper) (caddyhttp.MiddlewareHandler, err
 
 	switch args[0] {
 	case "strip_prefix":
-		if len(args) > 2 {
+		if len(args) != 2 {
 			return nil, h.ArgErr()
 		}
 		rewr.StripPathPrefix = args[1]
-		if !strings.HasPrefix(rewr.StripPathPrefix, "/") {
-			rewr.StripPathPrefix = "/" + rewr.StripPathPrefix
-		}
 
 	case "strip_suffix":
-		if len(args) > 2 {
+		if len(args) != 2 {
 			return nil, h.ArgErr()
 		}
 		rewr.StripPathSuffix = args[1]
